<?php
/**
 * Main plugin class definition and instantiation.
 *
 * @package ACWPB
 */

/**
 * Main plugin class.
 *
 * Handles all integration with WordPress an WPBackery Page Builder.
 */
class rrj_charts {

	/**
	 * The unique instance.
	 *
	 * @var rrj_charts
	 */
	private static $instance = null;

	/**
	 * Plugin option name.
	 */
	const options = 'rrj-charts-settings';

	/**
	 * PHP function option name.
	 */
	const php_functions = 'rrj-php-func';

    /**
     * Association of function names and files.
     */
    const PHP_FILENAMES = 'rrj-php-names';
    
	/**
	 * Default values for front end conditional variables.
	 *
	 * @var array
	 */
	private static $default_conditionals = array(
		'is_front_page'     => false,
		'is_home'           => false,
		'is_singular'       => false,
		'is_page'           => false,
		'is_user_logged_in' => false,
	);

	/**
	 * Chart settings page hook.
	 *
	 * @var string
	 */
	private $page_hook;

	/**
	 * Default plugin option.
	 *
	 * @var array
	 */
	private static $default_options;

	/**
	 * Unique instance constructor.
	 */
	private function __construct() {
		if ( defined( 'WPB_VC_VERSION' ) ) {
			add_action( 'init', array( $this, 'init' ) );

			add_action( 'plugins_loaded', array( $this, 'plugins_loaded' ) );

			// load all chart elements.
			require_once RRJ_AC_PATH . 'elements/line/line-chart.php';
			require_once RRJ_AC_PATH . 'elements/bar/bar-chart.php';
			require_once RRJ_AC_PATH . 'elements/bubble/bubble-chart.php';
			require_once RRJ_AC_PATH . 'elements/pie/pie-chart.php';
			require_once RRJ_AC_PATH . 'elements/radar/radar-chart.php';
			require_once RRJ_AC_PATH . 'elements/polar/polar-chart.php';

			// page inline scripts.
			add_action( 'wp_head', array( $this, 'wp_head' ) );

			add_action( 'admin_menu', array( $this, 'admin_menu' ) );

			// front end scripts.
			add_action( 'wp_enqueue_scripts', array( $this, 'enqueue_scripts' ) );

			// back end scripts.
			add_action( 'admin_enqueue_scripts', array( $this, 'admin_scripts' ) );

			// back end inline scripts.
			add_action( 'admin_footer', array( $this, 'admin_footer' ) );

			add_action( 'wp_ajax_rrj_csv_importer', array( $this, 'ajax_csv_import' ) );
			add_action( 'wp_ajax_rrj_save_settings', array( $this, 'ajax_save_settings' ) );
			add_action( 'wp_ajax_rrj_save_function', array( $this, 'ajax_save_function' ) );
			add_action( 'wp_ajax_rrj_delete_function', array( $this, 'ajax_delete_function' ) );
			add_action( 'wp_ajax_rrj_reset_settings', array( $this, 'ajax_reset_settings' ) );
			add_action( 'wp_ajax_rrj_new_fn', array( $this, 'ajax_create_function' ) );
			add_action( 'wp_ajax_rrj_test_fn', array( $this, 'ajax_test_function' ) );
			add_action( 'wp_ajax_rrj_upgrade_fn', array( $this, 'ajax_upgrade_function' ) );

			add_action( 'wp_ajax_rrj_run_fn', array( $this, 'ajax_run_function' ) );
			add_action( 'wp_ajax_nopriv_rrj_run_fn', array( $this, 'ajax_run_function' ) );
			add_action( 'rest_api_init', array( $this, 'register_route' ) );

			// meta box.
			add_action( 'add_meta_boxes', array( $this, 'meta_box' ), 10, 2 );
			add_action( 'save_post', array( $this, 'save_post' ), 10 );
		} else {
			add_action( 'admin_notices', array( $this, 'missing_wbpb' ) );
		}
	}

	/**
	 * Update post meta from meta box.
	 *
	 * @param integer $post_id The current post ID.
	 */
	public function save_post( $post_id ) {
		if ( ! empty( $_POST['rrjac']['enabled'] ) ) {

			$value = max( 0, min( 100, absint( $_POST['rrjac']['value'] ) ) );
			update_post_meta(
				$post_id,
				RRJ_AC_METAINIT,
				array(
					'enabled' => true,
					'value'   => $value,
				)
			);

		} else {
			delete_post_meta( $post_id, RRJ_AC_METAINIT );
		}
	}

	/**
	 * Add chart initialization meta box.
	 *
	 * @param string  $post_type The current post type.
	 * @param WP_POST $post The current Post.
	 */
	public function meta_box( $post_type, $post ) {
		$post_types = get_post_types( array( 'show_ui' => true ) );
		add_meta_box(
			'rrj-ac-init',
			__( 'Chart Initialization', 'rrj-ac' ),
			array( $this, 'print_meta_box' ),
			$post_types,
			'side',
			'default'
		);
	}

	/**
	 * Print the meta box markup.
	 *
	 * @param WP_POST $post The current post.
	 * @param array   $args Meta box arguments.
	 */
	public function print_meta_box( $post, $args ) {
		require_once RRJ_AC_PATH . 'includes/metabox.php';
	}

	/**
	 * Format imported data sets.
	 *
	 * @param string $type Chart type.
	 * @param array  $_sets The sets to be formated.
	 * @param array  $sets The original sets variable.
	 *
	 * @return array $sets The formated sets.
	 */
	private function format_sets( $type, $_sets, &$sets ) {

		if ( 'bubble' === $type ) {

			$__sets = array();
			foreach ( $_sets as $i => $set ) {

				$__sets[ $i ] = array();
				$point        = array();
				foreach ( $set as $cell ) {
					$point[] = $cell;
					if ( 3 === count( $point ) ) {
						$__sets[ $i ][] = '[' . implode( ';', $point ) . ']';
						$point          = array();
					}
				}
				if ( ! empty( $point ) ) {
					$__sets[ $i ][] = '[' . implode( ';', $point ) . ']';
				}
				$sets[] = implode( ' ', $__sets[ $i ] );

			}
		} else {

			foreach ( $_sets as $set ) {
				$sets[] = implode( ';', $set );
			}
		}
		return $sets;

	}

	/**
	 * Format the values imported from a CSV file to match the default dataset format (semicolon separate values in general).
	 *
	 * @param string $content the file content.
	 * @param string $type The chart type.
	 * @param string $orientation whether the sets are arranged vertically or horizontally.
	 * @param string $header Whether to skip the 1st row or column or use it as labels.
	 *
	 * @return type The formated sets.
	 */
	private function format_content( $content, $type, $orientation, $header ) {
		$labels = '';
		$sets   = array();

		if ( 'rows' === $orientation ) {

			$_sets = array();

			foreach ( $content as $i => $row ) {

				if ( 0 === $i ) {
					if ( 'skip' === $header ) {
						continue;
					}
					if ( '1st' === $header ) {
						$labels = str_replace( ',', ';', trim( $row ) );
						continue;
					}
				}
				$_sets[] = explode( ',', $row );

			}
			$this->format_sets( $type, $_sets, $sets );

		} else {

			$_labels = array();
			$_sets   = array();

			foreach ( $content as $i => $row ) {

				$cells = explode( ',', $row );
				foreach ( $cells as $j => $cell ) {

					$_j = $j;
					if ( 0 === $j ) {
						if ( 'skip' === $header ) {
							continue;
						}
						if ( '1st' === $header ) {
							$_labels[] = $cell;
							continue;
						}
					}
					if ( 'none' !== $header ) {
						$_j = $j - 1;
					}
					if ( ! isset( $_sets[ $_j ] ) ) {
						$_sets[ $_j ] = array();
					}
					$_sets[ $_j ][] = $cell;

				}
			}

			$labels = implode( ';', $_labels );
			$this->format_sets( $type, $_sets, $sets );

		}

		return array(
			'labels' => $labels,
			'sets'   => $sets,
		);
	}

	/**
	 * Run (eval) a PHP code.
	 *
	 * @param string $code PHP code.
	 * @param array  $conditionals Front end conditional variables.
	 * @param mixed  $stored_value Any previously stored return value.
	 *
     * @deprecated since v1.6 PHP code now written in files within the upload folder.
     * 
	 * @return mixed the value returned by the function.
	 */
	private function _eval( $code, $conditionals = array(), $stored_value = '' ) {

		$functions = get_option( self::php_functions, array() );
		if ( empty( $conditionals ) ) {
			$conditionals = self::$default_conditionals;
		}

		$code = trim( $code );

		if ( 0 === strpos( $code, '<?php' ) ) {
			$code = substr( $code, 5 );
		} elseif ( 0 === strpos( $code, '<?' ) ) {
			$code = substr( $code, 2 );
		}

		$results = eval( $code );
		return $results;

	}

	/**
	 * Register the REST API route.
	 */
	public function register_route() {

		register_rest_route(
			'rrj-advanced-charts/v1',
			'/function/(?P<fn>.+)/(?P<cond>.+)',
			array(
				'methods'  => 'GET',
				'callback' => array( $this, 'api_run_fn' ),
			)
		);

	}

	/**
	 * Run a functino called from the REST API.
	 *
	 * @param WP_REST_Request $request The REST API request.
	 *
	 * @return string The JSON response.
	 */
	public function api_run_fn( $request ) {
		$fn           = $request->get_param( 'fn' );
		$conds        = $request->get_param( 'cond' );
		$conditionals = json_decode( urldecode( $conds ), true );

		$results = $this->run_function( $fn, $conditionals );

		return rest_ensure_response( $results );
	}

	/**
	 * Execute a function or return the cached value (and eventually update the cache).
	 *
	 * @param string $fn The function name.
	 * @param array  $conditionals Conditional variables from the front end.
	 *
	 * @return mixed The return value of the user function.
	 */
	private function run_function( $fn, $conditionals ) {
		$functions = get_option( self::php_functions, array() );

		if ( isset( $functions[ $fn ] ) ) {

			$interval = array(
				'5min'  => 300,
				'15min' => 900,
				'30min' => 1800,
				'1h'    => 3600,
				'2h'    => 7200,
				'3h'    => 10800,
				'6h'    => 21600,
				'12h'   => 43200,
				'24h'   => 86400,
				'2d'    => 172800,
			);

			if ( 'disabled' !== $functions[ $fn ]['storage'] ) {
                
                // Cache still valid.
				if ( $functions[ $fn ]['lastRes'] && time() < $functions[ $fn ]['lastExec'] + $interval[ $functions[ $fn ]['storage'] ] ) {

					return $functions[ $fn ]['lastRes'];

				} else {

                    if ( isset( $functions[ $fn ]['file'] ) ) {
                        
                        $wud = wp_upload_dir( null, false );
                        $updir = str_replace( '\\', '/', trailingslashit( $wud['basedir'] ) . 'rrj-advanced-charts' );
                        
                        $stored_value = $functions[ $fn ]['lastRes'];
                        $results = require $updir . '/' . $functions[ $fn ]['file'];
                        
                    } else {
                        $results = $this->_eval( $functions[ $fn ]['code'], $conditionals, $functions[ $fn ]['lastRes'] );
                    }
                    
					$functions[ $fn ]['lastRes']  = $results;
					$functions[ $fn ]['lastExec'] = time();

					update_option( self::php_functions, $functions );
					return $results;

				}
                
			} else {
                // No cache - evaluate every time.
                if ( isset( $functions[ $fn ]['file'] ) ) {
                    
                    $wud = wp_upload_dir( null, false );
                    $updir = str_replace( '\\', '/', trailingslashit( $wud['basedir'] ) . 'rrj-advanced-charts' );
                    
                    $stored_value = $functions[ $fn ]['lastRes'];
                    $results = require $updir . '/' . $functions[ $fn ]['file'];
                    
                } else {
                    $results = $this->_eval( $functions[ $fn ]['code'], $conditionals, $functions[ $fn ]['lastRes'] );
                }
                
				return $results;

			}
		}
	}

	/**
	 * Run a function requested from AJAX.
	 */
	public function ajax_run_function() {
		$fn           = isset( $_POST['fn'] ) ? wp_unslash( $_POST['fn'] ) : false;
        $conditionals = array();
		if ( isset( $_POST['conditionals'] ) ) {
			try {
				$conditionals = json_decode( wp_unslash( $_POST['conditionals'] ), true );
			} catch ( Exception $e ) {
                $conditionals = self::$default_conditionals;
			}
		}
        
        $conditionals = $conditionals + self::$default_conditionals;
        
		if ( $fn ) {

			$results = $this->run_function( $fn, $conditionals );
			header( 'Content-Type: text/html' );
			echo wp_json_encode( $results );
		}
		die;
	}

	/**
	 * Test a function in the back end.
	 */
	public function ajax_test_function() {
        
        $_post = wp_unslash( $_POST );
		$nonce = isset( $_post['nonce'] ) ? $_post['nonce'] : '';

		if ( false !== wp_verify_nonce( $nonce, 'rrj-settings' ) ) {
			try {
				$conditionals = json_decode( $_post['conditionals'], true );
			} catch ( Exception $e ) {
                $conditionals = array();
			}
            
            $fn = $_post['fn'];
			$functions = get_option( self::php_functions, array() );
            
            $wud = wp_upload_dir( null, false );
            $updir = str_replace( '\\', '/', trailingslashit( $wud['basedir'] ) . 'rrj-advanced-charts' );
            
            $results = 'the for function "' . $fn . '" does not exists';
            
            if ( file_exists( $updir . '/' . $functions[ $fn ]['file'] ) ) {
                $stored_value = null;
                $results = require $updir . '/' . $functions[ $fn ]['file'];
            }
            
			header( 'Content-Type: text/html' );
			echo wp_json_encode( $results );
		}
		die;

	}

	/**
	 * Delete a function and all its parameters.
	 */
	public function ajax_delete_function() {

        $_post = wp_unslash( $_POST );
		$nonce = isset( $_post['rrj-settings-nonce'] ) ? wp_unslash( $_post['rrj-settings-nonce'] ) : '';

		if ( false !== wp_verify_nonce( $nonce, 'rrj-settings' ) ) {

			$fn        = $_post['fn'];
			$functions = get_option( self::php_functions, array() );

            if ( isset( $functions[ $fn ] ) ) {
                if ( isset( $functions[ $fn ]['file'] ) ) {
                    
                    $fname = $functions[ $fn ]['file'];
                    $fs = Rrj_Filesystem::get_instance();
                    $creds = $fs->check_creds();
                    
                    if ( $creds ) {
                        
                        if ( WP_Filesystem( $creds ) ) {
                            
                            global $wp_filesystem;
                            $_wud = wp_upload_dir( null, false );
                            $_updir = str_replace( '\\', '/', trailingslashit( $_wud['basedir'] ) . 'rrj-advanced-charts' );
                            
                            $updir = $fs->normalize_path( $_updir );
                            
                            if ( ! defined( 'FS_CHMOD_FILE' ) ) {
                                define( 'FS_CHMOD_FILE', ( 0644 & ~ umask() ) );
                            }
                            if ( ! defined( 'FS_CHMOD_DIR' ) ) {
                                define( 'FS_CHMOD_DIR', ( 0755 & ~ umask() ) );
                            }
                            
                            if ( file_exists( $_updir . '/' . $fname ) ) {
                                
                                if ( $wp_filesystem->delete( $updir . '/' . $fname ) ) {
                                    
                                    unset( $functions[ $fn ] );
                                    update_option( self::php_functions, $functions );
                                    
                                    header( 'Content-type: application/json' );
                                    echo wp_json_encode( array(
                                        'status' => true,
                                        'action' => 'deleteFunction',
                                    ) );
                                    
                                } else {
                                    header( 'Content-type: application/json' );
                                    echo wp_json_encode( array(
                                        'status' => false,
                                        'endError' => sprintf( __( 'Can not delete the file "%s"', 'rrj-ac' ),$updir . '/' . $fname  ),
                                    ) );
                                    die;
                                }
                                
                            } else {
                                
                                unset( $functions[ $fn ] );
                                update_option( self::php_functions, $functions );
                                
                                header( 'Content-type: application/json' );
                                echo wp_json_encode( array(
                                    'status' => true,
                                    'action' => 'deleteFunction',
                                ) );
                            }
                            
                        } else {
                            
                            $fields = array( 'action', 'fn', 'rrj-settings-nonce' );
                            $form = $fs->get_form( admin_url( 'admin-ajax.php' ), $fields, true );
                            header( 'Content-type: application/json' );
                            echo wp_json_encode( array(
                                'status' => false,
                                'form' => $form,
                            ) );
                            
                        }
                        
                    } else {
                        
                        $fields = array( 'action', 'fn', 'rrj-settings-nonce' );
                        $error = ( isset( $_post['_fs_nonce'] ) );
                        $form = $fs->get_form( admin_url( 'admin-ajax.php' ), $fields, $error );
                        header( 'Content-type: application/json' );
                        echo wp_json_encode( array(
                            'status' => false,
                            'form' => $form,
                        ) );
                        
                    }
                    
                } else {
                    unset( $functions[ $fn ] );
                    update_option( self::php_functions, $functions );
                }
            }

		}
		die;

	}

	/**
	 * Save a function and all its parameters.
	 */
	public function ajax_save_function() {

        $_post = wp_unslash( $_POST );
		$nonce = isset( $_post['rrj-settings-nonce'] ) ? $_post['rrj-settings-nonce'] : '';
		if ( false !== wp_verify_nonce( $nonce, 'rrj-settings' ) ) {

			$code    = $_post['code'];
			$fn      = $_post['fn'];
			$storage = $_post['storage'];

			$functions = get_option( self::php_functions, array() );
            
            $fs = Rrj_Filesystem::get_instance();
            $creds = $fs->check_creds();
            
            if ( $creds ) {
                if ( WP_Filesystem( $creds ) ) {
                    
                    global $wp_filesystem;
                    $_wud = wp_upload_dir( null, false );
                    $_updir = str_replace( '\\', '/', trailingslashit( $_wud['basedir'] ) . 'rrj-advanced-charts' );
                    
                    $updir = $fs->normalize_path( $_updir );
                    
                    if ( ! defined( 'FS_CHMOD_FILE' ) ) {
                        define( 'FS_CHMOD_FILE', ( 0644 & ~ umask() ) );
                    }
                    if ( ! defined( 'FS_CHMOD_DIR' ) ) {
                        define( 'FS_CHMOD_DIR', ( 0755 & ~ umask() ) );
                    }
                    
                    if ( !file_exists( $_updir ) || ! is_dir( $_updir ) ) {
                        if ( ! $wp_filesystem->mkdir( $updir, FS_CHMOD_DIR ) ) {
                            
                            header( 'Content-type: application/json' );
                            echo wp_json_encode( array(
                                'status' => false,
                                'endError' => sprintf( __( 'Can not create the directory "%s"', 'rrj-ac' ),$_updir ),
                            ) );
                            die;
                            
                        }
                    }
                    
                    $fname = '';
                    
                    if ( isset( $functions[ $fn ]['file'] ) && '' !== $functions[ $fn ]['file'] ) {
                        $fname = $functions[ $fn ]['file'];
                    } else {
                        $fname = wp_generate_password( 8, false );
                        $fname .= '.php';
                    }
                    
                    if ( 0 !== strpos( $code, '<?php' ) ) {
                        $code = "<?php\n" . $code; 
                    }
                    
                    if ( $wp_filesystem->put_contents( $updir . '/' . $fname, $code, FS_CHMOD_FILE ) ) {
                        
                        /**
                         * Store old value to preserve last execution time and results.
                         */
                        $old_func  = $functions[ $fn ];

                        /**
                         * Remove the code field.
                         */
                        if ( isset( $old_func['code'] ) ) {
                            unset( $old_func['code'] );
                        }
                        
                        $functions[ $fn ] = array(
                            'file' => $fname,
                            'storage' => $storage,
                        ) + $old_func;
                        
                        update_option( self::php_functions, $functions );
                        
                        header( 'Content-type: application/json' );
                        echo wp_json_encode( array(
                            'status' => true,
                            'fileContent' => $code, 
                            'action' => 'saveFunction',
                        ) );
                        
                    } else {
                        
                        header( 'Content-type: application/json' );
                        echo wp_json_encode( array(
                            'status' => false,
                            'endError' => sprintf( __( 'Can not write the function "%s" into the file "%s"', 'rrj-ac' ), esc_html( $fn ), $_updir . '/' . $fname ),
                        ) );
                        
                    }
                    
                } else {
                    
                    $fields = array( 'code', 'fn', 'storage', 'rrj-settings-nonce' );
                    $form = $fs->get_form( admin_url( 'admin-ajax.php?action=rrj_save_function' ), $fields, true );
                    header( 'Content-type: application/json' );
                    echo wp_json_encode( array(
                        'status' => false,
                        'form' => $form,
                    ) );
                }
                
            } else {
                
                /**
                 * The fields to include into the credentials form.
                 */
                $fields = array( 'code', 'fn', 'storage', 'rrj-settings-nonce' );
                
                $error = ( isset( $_post['_fs_nonce'] ) );
                $form = $fs->get_form( admin_url( 'admin-ajax.php?action=rrj_save_function' ), $fields, $error );
                
                header( 'Content-Type: application/json' );
                echo wp_json_encode( array(
                    'status' => false,
                    'form' => $form,
                ) );
                
            }
            
		}
		die;

	}

	/**
	 * Move functions from options into solid PHP files.
	 */
	public function ajax_upgrade_function() {

        $_post = wp_unslash( $_POST );
		$nonce = isset( $_post['rrj-settings-nonce'] ) ? $_post['rrj-settings-nonce'] : '';
		if ( false !== wp_verify_nonce( $nonce, 'rrj-settings' ) ) {

			$functions = get_option( self::php_functions, array() );
            
            $fs = Rrj_Filesystem::get_instance();
            $creds = $fs->check_creds();
            
            if ( $creds ) {
                if ( WP_Filesystem( $creds ) ) {
                    
                    global $wp_filesystem;
                    $_wud = wp_upload_dir( null, false );
                    $_updir = str_replace( '\\', '/', trailingslashit( $_wud['basedir'] ) . 'rrj-advanced-charts' );
                    
                    $updir = $fs->normalize_path( $_updir );
                    
                    if ( ! defined( 'FS_CHMOD_FILE' ) ) {
                        define( 'FS_CHMOD_FILE', ( 0644 & ~ umask() ) );
                    }
                    if ( ! defined( 'FS_CHMOD_DIR' ) ) {
                        define( 'FS_CHMOD_DIR', ( 0755 & ~ umask() ) );
                    }
                    
                    if ( !file_exists( $_updir ) || ! is_dir( $_updir ) ) {
                        if ( ! $wp_filesystem->mkdir( $updir, FS_CHMOD_DIR ) ) {
                            
                            header( 'Content-type: application/json' );
                            echo wp_json_encode( array(
                                'status' => false,
                                'endError' => sprintf( __( 'Can not create the directory "%s"', 'rrj-ac' ),$_updir ),
                            ) );
                            die;
                            
                        }
                    }
                    
                    foreach ( $functions as $fn => $value ) {
                        if ( isset( $value['code'] ) ) {
                            $code = $value['code'];
                            $storage = $value['storage'];
                            $fname = wp_generate_password( 8, false );
                            $fname .= '.php';
                            while ( file_exists( $_updir . '/' . $fname ) ) {
                                $fname = wp_generate_password( 8, false );
                                $fname .= '.php';
                            }
                            
                            if ( 0 !== strpos( $code, '<?php' ) ) {
                                $code = "<?php\n" . $code; 
                            }
                            
                            if ( $wp_filesystem->put_contents( $updir . '/' . $fname, $code, FS_CHMOD_FILE ) ) {
                                
                                /**
                                 * Store old value to preserve last execution time and results.
                                 */
                                $old_func  = $functions[ $fn ];

                                /**
                                 * Remove the code field.
                                 */
                                if ( isset( $old_func['code'] ) ) {
                                    unset( $old_func['code'] );
                                }
                                
                                $functions[ $fn ] = array(
                                    'file' => $fname,
                                    'storage' => $storage,
                                ) + $old_func;
                                
                                update_option( self::php_functions, $functions );
                                
                            } else {
                                
                                header( 'Content-type: application/json' );
                                echo wp_json_encode( array(
                                    'status' => false,
                                    'endError' => sprintf( __( 'Can not write the function "%s" into the file "%s"', 'rrj-ac' ), esc_html( $fn ), $_updir . '/' . $fname ),
                                ) );
                                die;
                            }
                        }
                    }
                    
                    header( 'Content-type: application/json' );
                    echo wp_json_encode( array(
                        'status' => true,
                        'action' => 'upgradeFunctions',
                    ) );
                    die;
                    
                } else {
                    
                    $fields = array( 'action', 'rrj-settings-nonce' );
                    $form = $fs->get_form( admin_url( 'admin-ajax.php?action=rrj_run_fn' ), $fields, true );
                    header( 'Content-type: application/json' );
                    echo wp_json_encode( array(
                        'status' => false,
                        'form' => $form,
                    ) );
                }
                
            } else {
                
                /**
                 * The fields to include into the credentials form.
                 */
                $fields = array( 'action', 'rrj-settings-nonce' );
                
                $error = ( isset( $_post['_fs_nonce'] ) );
                $form = $fs->get_form( admin_url( 'admin-ajax.php' ), $fields, $error );
                
                header( 'Content-Type: application/json' );
                echo wp_json_encode( array(
                    'status' => false,
                    'form' => $form,
                ) );
                
            }
            
		}
		die;

	}

	/**
	 * Add a new PHP function to the list.
	 */
	public function ajax_create_function() {

		$nonce = isset( $_POST['nonce'] ) ? wp_unslash( $_POST['nonce'] ) : '';
		if ( false !== wp_verify_nonce( $nonce, 'rrj-settings' ) ) {

			$fn = wp_unslash( $_POST['fn'] );

			if ( $fn ) {

				$functions        = get_option( self::php_functions, array() );
				$functions[ $fn ] = array(
					'file'     => '',
					'storage'  => '3h',
					'lastExec' => 0,
					'lastRes'  => '',
				);
				update_option( self::php_functions, $functions, false );
				$result = array(
					'status'    => true,
					'functions' => $functions,
				);

			} else {
				$result = array( 'status' => false );
			}

			header( 'Content-Type: application/json' );
			echo wp_json_encode( $result );

		}
		die;

	}

	/**
	 * Reset plugin settings.
	 */
	public function ajax_reset_settings() {

		$nonce = isset( $_POST['nonce'] ) ? wp_unslash( $_POST['nonce'] ) : '';
		if ( false !== wp_verify_nonce( $nonce, 'rrj-settings' ) ) {

			update_option( self::options, self::$default_options );

			$result = array( 'status' => true );

			header( 'Content-Type: application/json' );
			echo wp_json_encode( $result );

		}
		die;

	}

	/**
	 * Save plugin settings (AJAX).
	 */
	public function ajax_save_settings() {

		$nonce = isset( $_POST['nonce'] ) ? wp_unslash( $_POST['nonce'] ) : '';

		if ( false !== wp_verify_nonce( $nonce, 'rrj-settings' ) ) {

			parse_str( wp_unslash( $_POST['args'] ), $args );
			$args = wp_unslash( $args );

			$force_axes_color  = isset( $args['force-axes-color'] );
			$axes_color        = sanitize_hex_color( $args['axes-color'] ) ? sanitize_hex_color( $args['axes-color'] ) : '#646464';
			$legend_font_color = sanitize_hex_color( $args['legend-font-color'] ) ? sanitize_hex_color( $args['legend-font-color'] ) : '#646464';

			$options = array(
				'force-axes-color'  => $force_axes_color,
				'axes-color'        => $axes_color,
				'legend-font-color' => $legend_font_color,
			) +
			$args +
			self::$default_options;

			update_option( self::options, $options );

			$result = array( 'status' => true );

			header( 'Content-Type: aplication/json' );
			echo wp_json_encode( $result );

		}
		die;

	}

	/**
	 * Ajax callback for data import.
	 */
	public function ajax_csv_import() {

		$nonce = isset( $_POST['nonce'] ) ? wp_unslash( $_POST['nonce'] ) : '';

		if ( false !== wp_verify_nonce( $nonce, 'rrj-importer' ) ) {

			$id          = absint( $_POST['id'] );
			$type        = wp_unslash( $_POST['type'] );
			$orientation = wp_unslash( $_POST['orientation'] );
			$header      = wp_unslash( $_POST['header'] );
			$file        = get_attached_file( $id );

			$result = $this->parse_csv(
				$file,
				array(
					'type'        => $type,
					'orientation' => $orientation,
					'header'      => $header,
				)
			);

			header( 'Content-Type: application/json' );
			echo wp_json_encode(
				array(
					'status' => true,
					'data'   => $result,
				)
			);

		}
		die;

	}

	/**
	 * Fill short rows with empty values.
	 *
	 * @param array   $rows Array of rows.
	 * @param integer $max the maximum row length (for recursive calls).
	 *
	 * @return array Rows with equalized length.
	 */
	private function normalize_rows( $rows = array(), $max = 0 ) {
		$repeat = false;
		foreach ( $rows as $i => $row ) {
			$count = substr_count( $row, ',' );
			if ( $count > $max ) {
				$max    = $count;
				$repeat = true;
				break;
			} elseif ( $count < $max ) {
				$rows[ $i ] .= str_repeat( ',', $max - $count );
			}
		}
		if ( $repeat ) {
			return $this->normalize_rows( $rows, $max );
		} else {
			return $rows;
		}
	}

	/**
	 * Parse CSV files.
	 *
	 * @param string $file Filename with full path.
	 * @param array  $options Parsing options.
	 *
	 * @return array Formated content (labels and sets).
	 */
	public function parse_csv( $file, $options = array() ) {

		$default_options = array(
			'type'        => 'line',
			'orientation' => 'rows',
			'header'      => 'skip',
		);

		$options += $default_options;

		$content = file_get_contents( $file );
		$content = explode( "\n", $content );
		$content = $this->normalize_rows( $content );
		$content = $this->format_content( $content, $options['type'], $options['orientation'], $options['header'] );

		return $content;

	}

	/**
	 * Admin inline scripts and markup.
	 */
	public function admin_footer() {
		$scr        = get_current_screen();
		$vc_screens = array( 'post', 'edit', 'widgets' );

		if ( in_array( $scr->base, $vc_screens, true ) ) {
			require_once RRJ_AC_PATH . 'assets/css/admin-footer.css.php';
			echo '<div style="display:none;" id="rrj-load-overlay"><img alt="loading" src="' . esc_url( RRJ_AC_URL . 'assets/images/loading.gif' ) . '" /></div>';
		}
	}

	/**
	 * "wp_head"
	 */
	public function wp_head() {
		$options = $this->get_option();

		/**
		 * Overwrite init option
		 */
		if ( is_singular() ) {
			global $post;
			if ( $post ) {
				$meta = get_post_meta( $post->ID, RRJ_AC_METAINIT, true );
				if ( is_array( $meta ) && $meta['enabled'] ) {
					$options['init'] = $meta['value'];
				}
			}
		}

		echo '<script type="text/javascript"> var rrjChartPluginOptions = ' . wp_json_encode( $options ) . ';</script>';

	}

	/**
	 * Add chart settings page.
	 */
	public function admin_menu() {
		$this->page_hook = add_submenu_page(
			'vc-general',
			__( 'Charts settings', 'rrj-ac' ),
			__( 'Charts settings', 'rrj-ac' ),
			'manage_options',
			'rrj_charts',
			array( $this, 'settings_page' )
		);
	}

	/**
	 * Settings page callback function.
	 */
	public function settings_page() {
		require_once RRJ_AC_PATH . 'includes/settings-page.php';
	}

	/**
	 * Enqueue back end scripts.
	 */
	public function admin_scripts() {
		$scr        = get_current_screen();
		$vc_screens = array( 'post', 'edit', 'widgets' );

		/**
		 * VC Screens
		 */
		if ( in_array( $scr->base, $vc_screens, true ) ) {

			wp_enqueue_script( 'rrj-codemirror', RRJ_AC_URL . 'assets/libs/CodeMirror/codemirror.js', array( 'jquery' ), RRJ_AC_VERSION );
			wp_enqueue_script( 'rrj-jsmode', RRJ_AC_URL . 'assets/libs/CodeMirror/javascript.js', array( 'rrj-codemirror' ), RRJ_AC_VERSION );
			wp_enqueue_script( 'rrj-jseditor', RRJ_AC_URL . 'assets/js/param-jseditor.js', array( 'rrj-codemirror' ), RRJ_AC_VERSION );
			wp_enqueue_script( 'rrj-mediaframe', RRJ_AC_URL . 'assets/js/media-frame.js', array( 'jquery' ), RRJ_AC_VERSION );

			$media_locale = array(
				'selectMedias'    => __( 'Select files', 'rrj-ac' ),
				'selectMedia'     => __( 'Select a file', 'rrj-ac' ),
				'button'          => __( 'select', 'rrj-ac' ),
				'invalidFileType' => __( 'invalid file type', 'rrj-ac' ),
			);
			wp_register_script( 'rrj-importer', RRJ_AC_URL . 'assets/js/param-importer.js', array( 'rrj-mediaframe' ), RRJ_AC_VERSION );
			wp_localize_script( 'rrj-importer', 'rrjMediaFrameLocale', $media_locale );
			wp_enqueue_script( 'rrj-importer' );
			wp_enqueue_style( 'rrj-codemirror', RRJ_AC_URL . 'assets/libs/CodeMirror/codemirror.css', array(), RRJ_AC_VERSION );

		}

		/**
		 * Settings page
		 */
		if ( $scr->id == $this->page_hook ) {
            Rrj_Filesystem::enqueue_rfc();
			wp_enqueue_style( 'rrj-settings', RRJ_AC_URL . 'assets/css/admin-settings.css', array(), RRJ_AC_VERSION );
			wp_enqueue_style( 'wp-color-picker' );
			wp_enqueue_style( 'rrj-codemirror', RRJ_AC_URL . 'assets/libs/CodeMirror/codemirror.css', array(), RRJ_AC_VERSION );

			wp_enqueue_script( 'rrj-codemirror', RRJ_AC_URL . 'assets/libs/CodeMirror/codemirror.js', array( 'jquery' ), RRJ_AC_VERSION );
			wp_enqueue_script( 'rrj-jsmode', RRJ_AC_URL . 'assets/libs/CodeMirror/javascript.js', array( 'rrj-codemirror' ), RRJ_AC_VERSION );
			wp_enqueue_script( 'rrj-phpmode', RRJ_AC_URL . 'assets/libs/CodeMirror/php.js', array( 'rrj-codemirror' ), RRJ_AC_VERSION );
			wp_enqueue_script( 'rrj-xmlmode', RRJ_AC_URL . 'assets/libs/CodeMirror/xml.js', array( 'rrj-codemirror' ), RRJ_AC_VERSION );
			wp_enqueue_script( 'rrj-cssmode', RRJ_AC_URL . 'assets/libs/CodeMirror/css.js', array( 'rrj-codemirror' ), RRJ_AC_VERSION );
			wp_enqueue_script( 'rrj-clikemode', RRJ_AC_URL . 'assets/libs/CodeMirror/clike.js', array( 'rrj-codemirror' ), RRJ_AC_VERSION );
			wp_enqueue_script( 'rrj-htmlmixedmode', RRJ_AC_URL . 'assets/libs/CodeMirror/htmlmixed.js', array( 'rrj-codemirror' ), RRJ_AC_VERSION );
			wp_enqueue_script( 'rrj-settings', RRJ_AC_URL . 'assets/js/admin-settings.js', array( 'jquery', 'wp-color-picker', 'rrj-codemirror' ), RRJ_AC_VERSION );

		}

	}

	/**
	 * Load text domain for translations.
	 */
	public function plugins_loaded() {
		load_plugin_textdomain( 'rrj-ac', false, RRJ_AC_PATH . 'languages' );
	}

	/**
	 * Initialization.
	 */
	public function init() {
		// add rrj_uislider param type.
		vc_add_shortcode_param( 'rrj_uislider', array( $this, 'uislider' ) );

		// add rrj_jseditor param type.
		vc_add_shortcode_param( 'rrj_jseditor', array( $this, 'jseditor' ) );

		// add rrj_importer param type.
		vc_add_shortcode_param( 'rrj_importer', array( $this, 'importer' ) );

		self::$default_options = array(
			'force-axes-color'  => false,
			'axes-color'        => '#3a3a3a',
			'legend-font-color' => '#3a3a3a',
			'legend-font-size'  => '12',
			'font-family'       => '-apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, "Helvetica Neue", Arial, sans-serif',
			't-separator'       => '',
			'init'              => '33',
		);
	}

    /**
     * "jseditor" parameter markup.
     *
     * @param array  $settings The parameter settings.
     * @param string $value The current value of the parameter.
     *
     * @return string The ouptut.
     */
	public function jseditor( $settings, $value ) {
		ob_start();
		include plugin_dir_path( __FILE__ ) . '/jseditor.php';
		return ob_get_clean();
	}

	/**
	 * "rrj_importer" parameter markup.
	 *
	 * @param array  $settings The parameter settings.
	 * @param string $value The current value of the parameter.
	 *
	 * @return string The ouptut.
	 */
	public function importer( $settings, $value ) {
		ob_start();
		include plugin_dir_path( __FILE__ ) . '/importer.php';
		return ob_get_clean();
	}

	/**
	 * "uislider" parameter markup.
	 *
	 * @param array  $settings The parameter settings.
	 * @param string $value The current value of the parameter.
	 *
	 * @return string The ouptut.
	 */
	public function uislider( $settings, $value ) {
		ob_start();
		include plugin_dir_path( __FILE__ ) . '/uislider.php';
		return ob_get_clean();
	}

	/**
	 * Enqueue scripts and style on the front end.
	 */
	public function enqueue_scripts() {

		$conditionals = array(
			'is_front_page'     => is_front_page(),
			'is_home'           => is_home(),
			'is_singular'       => is_singular(),
			'is_page'           => is_page(),
			'is_user_logged_in' => is_user_logged_in(),
			'post_id'           => get_the_id(),
		);
		$conditionals = wp_json_encode( $conditionals );
		$rest_url     = get_rest_url( null, 'rrj-advanced-charts/v1/function/' );

		$vars = array(
			'ajaxUrl'      => admin_url( 'admin-ajax.php' ),
			'restUrl'      => $rest_url,
			'condsJson'    => $conditionals,
			'condsEncoded' => rawurlencode( $conditionals ),
		);

		if ( defined( 'WP_DEBUG' ) && WP_DEBUG ) {

			wp_enqueue_script( 'rrj-charts', RRJ_AC_URL . 'assets/js/chart-controller.js', array( 'jquery' ), RRJ_AC_VERSION );
			wp_enqueue_script( 'rrj-data-loader', RRJ_AC_URL . 'assets/js/data-loaders.js', array( 'jquery' ), RRJ_AC_VERSION );

		} else {

			wp_enqueue_script( 'rrj-charts', RRJ_AC_URL . 'assets/js/chart-controller.min.js', array( 'jquery' ), RRJ_AC_VERSION );
			wp_enqueue_script( 'rrj-data-loader', RRJ_AC_URL . 'assets/js/data-loaders.min.js', array( 'jquery' ), RRJ_AC_VERSION );

		}
		wp_add_inline_script(
			'rrj-data-loader',
			';var rrjVars = ' . wp_json_encode( $vars ) . ';',
			'before'
		);

	}

	/**
	 * Add an admin notice about missing WPBakery Page Builder.
	 */
	public function missing_wbpb() {
		$msg = sprintf(
			__( '%1$s needs %2$s to be installed and activate to function', 'rrj-ac' ),
			'<strong>Advanced Charts</strong>',
			'<strong>WPBackery Page Builder</strong>'
		);
		echo '<div class="error"><p>' . wp_kses( $msg, array( 'strong' => true ) ) . '</p></div>';
	}

	/**
	 * Return the aspect ratio param mapping.
	 *
	 * @return array Array of ratios.
	 */
	public static function AR() {
		return array(
			'1:1'  => '1:1',
			'21:9' => '21:9',
			'16:9' => '16:9',
			'16:9' => '16:9',
			'4:3'  => '4:3',
			'3:4'  => '3:4',
			'9:16' => '9:16',
			'9:21' => '9:21',
		);
	}

	/**
	 * Process aspect ratio fields.
	 *
	 * @param string $ar aspect ratio.
	 * @param string $car Custom aspect ratio.
	 *
	 * @return array Array of width and height.
	 */
	public static function process_ar( $ar = '4:3', $car = '' ) {

		$AR  = explode( ':', $ar );
		$car = str_replace( ' ', '', $car );
		if ( ! empty( $car ) ) {
			$CAR = explode( ':', $car );
			if ( 2 === count( $CAR ) ) {
				if ( ! is_nan( floatval( $CAR[0] ) ) && ! empty( $CAR[0] ) && ! is_nan( floatval( $CAR[1] ) ) && ! empty( $CAR[1] ) ) {
					$AR = array( floatval( $CAR[0] ), floatval( $CAR[1] ) );
				}
			}
		}
		return $AR;

	}

	/**
	 * Get plugin's options
	 *
	 * @return array Plugin options.
	 */
	public function get_option() {
		$options  = get_option( self::options, array() );
		$options += self::$default_options;
		return $options;
	}

    /**
     * Get the code of a function
     */
    public static function get_function_code( $fn ) {
        $code = '';
        $functions = get_option( self::php_functions, array() );
        
        if ( isset( $functions[ $fn ] ) ) {
            $wud = wp_upload_dir( null, false );
            $updir = str_replace( '\\', '/', trailingslashit( $wud['basedir'] ) . 'rrj-advanced-charts' );
            
            if ( file_exists( $updir . '/' . $functions[ $fn ]['file'] ) ) {
                $code = file_get_contents( $updir . '/' . $functions[ $fn ]['file'] );
            }
        }
        
        return $code;
    }
    
	/**
	 * Return or create the unique instance.
	 *
	 * @return rrj_charts $instance.
	 */
	public static function instance() {
		if ( null === self::$instance ) {
			self::$instance = new self();
		}
		return self::$instance;
	}

}
rrj_charts::instance();
