<?php
/**
 * Color related tools.
 *
 * @package ACWPB
 */

if ( class_exists( 'rrj_colors', false ) ) {
	return;
}
/**
 * Color manipulation class.
 */
 
class rrj_colors {
    
	/**
	 * Instance of this class.
	 *
	 * @var object
	 */
	private static $instance = null;

	/**
	 * Constructs the unique instance of this class.
	 */
	private function __construct() {

	}

	/**
	 * Modifies (lower) the alpha channel (opacity) of a color.
	 *
	 * @param mixed $color the color in rgba, rgb or hex format.
	 * @param float $a the alpha channel factor.
	 * @param bool  $array whether to return the result in array (rgba) format.
	 *
	 * @throws Exception InvalidAlphaChannelFactor If the provided argument is not numeric.
	 */
	public function dilute( $color, $a = 0.5, $array = false ) {
		if ( ! is_numeric( $a ) ) {
			throw new Exception( 'InvalidAlphaChannelFactor' );
		}
		if ( 0 > $a ) {
			$a = 0;
		}
		if ( 1 < $a ) {
			$a = 1;
		}
		$c = $this->get_rgba( $color );
		$d = array(
			$c[0],
			$c[1],
			$c[2],
			$c[3] * $a,
		);
		return ( $array ) ? $d : $this->format_rgba( $d );
	}

	/**
	 * Get the valid CSS rgba color string.
	 *
	 * @param mixed $color the color.
	 */
	public function format_rgba( $color ) {
		$c = $this->get_rgba( $color );
		return 'rgba(' . intval( $c[0] ) . ',' . intval( $c[1] ) . ',' . intval( $c[2] ) . ',' . $c[3] . ')';
	}

	/**
	 * Get rgb array from HEX, rgba or rgb array.
	 *
	 * @param mixed $color the color.
	 */
	public function get_rgb( $color ) {
		$color = $this->get_rgba( $color );
		return array_slice( $color, 0, 3 );
	}

	/**
	 * Get rgba array from HEX, rgba or rgb array.
	 *
	 * @param mixed $color the color.
	 *
	 * @throws Exception InvalidColorString If the provided color is a valid HEX color.
	 * @throws Exception InvalidColorFormat If the format is not recognized (rgba or rgb in array from, or a valid HEX color with leading has tag).
	 */
	public function get_rgba( $color ) {
		if ( is_string( $color ) ) {
			$color = trim( strtolower( $color ) );
			if ( false !== strpos( $color, '#' ) ) {
				// hex color.
				return array_merge( $this->hex_to_rgb( $color ), array( 1 ) );
			}
			if ( false !== strpos( $color, 'rgba' ) ) {
				// check if rgba first.
				$color = str_replace( array( 'rgba', '(', ')', ' ' ), '', $color );
				$color = explode( ',', $color );
				if ( 4 > count( $color ) ) {
					$color[] = 1;
				}
				return array(
					floatval( $color[0] ),
					floatval( $color[1] ),
					floatval( $color[2] ),
					floatval( $color[3] ),
				);
			}
			if ( false !== strpos( $color, 'rgb' ) ) {
				// then check if rgb.
				$color = str_replace( array( 'rgb', '(', ')', ' ' ), '', $color );
				$color = explode( ',', $color );
				if ( 4 > count( $color ) ) {
					$color[] = 1;
				}
				return array(
					floatval( $color[0] ),
					floatval( $color[1] ),
					floatval( $color[2] ),
					floatval( $color[3] ),
				);
			}
			throw new Exception( 'InvalidColorString' );
		} elseif ( is_array( $color ) ) {
			// color in array form.
			$rgba   = array(
				floatval( $color[0] ),
				floatval( $color[1] ),
				floatval( $color[2] ),
			);
			$rgba[] = isset( $color[3] ) ? $color[3] : 1;
			return $rgba;
		} else {
			throw new Exception( 'InvalidColorFormat' );
		}
	}

	/**
	 * Convert HEX colors to rbg array.
	 *
	 * @param string $hex hexadecimal color (3 or 6 characters).
	 *
	 * @return array $rbg r g and b component of the color.
	 */
	public function hex_to_rgb( $hex ) {
		$hex = trim( str_replace( '#', '', $hex ) );

		if ( strlen( $hex ) === 3 ) {
			$r = hexdec( substr( $hex, 0, 1 ) . substr( $hex, 0, 1 ) );
			$g = hexdec( substr( $hex, 1, 1 ) . substr( $hex, 1, 1 ) );
			$b = hexdec( substr( $hex, 2, 1 ) . substr( $hex, 2, 1 ) );
		} else {
			$r = hexdec( substr( $hex, 0, 2 ) );
			$g = hexdec( substr( $hex, 2, 2 ) );
			$b = hexdec( substr( $hex, 4, 2 ) );
		}
		$rgb = array( $r, $g, $b );

		return $rgb;
	}

	/**
	 * Combine HEX color and apha channel (opacity) to get an rgba string.
	 *
	 * @param string $hex the color in HEX format.
	 * @param flot   $a the opacity value.
	 */
	public function hex_to_rgba( $hex, $a = 1 ) {
		$rgb = $this->get_rgb( $hex );
		return 'rgba(' . $rgb[0] . ',' . $rgb[1] . ',' . $rgb[2] . ',' . $a . ')';
	}

	/**
	 * Return color in the given format.
	 *
	 * @param mixed  $color the color.
	 * @param string $format the targeted format.
	 */
	public function format( $color, $format ) {
		$result = '';
		switch ( $format ) {
			case 'rgb':
				$color  = $this->get_rgb( $color );
				$result = 'rgb(' . $color[0] . ',' . $color[1] . ',' . $color[2] . ')';
				break;
			case 'hex':
				$color   = $this->get_rgb( $color );
				$hex_arr = array(
					( 1 === strlen( dechex( $color[0] ) ) ) ? '0' . dechex( $color[0] ) : dechex( $color[0] ),
					( 1 === strlen( dechex( $color[1] ) ) ) ? '0' . dechex( $color[1] ) : dechex( $color[1] ),
					( 1 === strlen( dechex( $color[2] ) ) ) ? '0' . dechex( $color[2] ) : dechex( $color[2] ),
				);
				$result  = '#' . implode( '', $hex_arr );
				break;
			case 'rgba':
				$result = $this->format_rgba( $color );
				break;
			default: // array.
				$result = $this->get_rgba( $color );
		}
		return $result;
	}

	/**
	 * Lighten color by a given amount.
	 *
	 * @param mixed  $color the color.
	 * @param float  $amount the lightening factor (0-100 ideally).
	 * @param string $format the format of the result.
	 */
	public function lighten( $color, $amount = 15, $format = 'array' ) {
		$rgba  = $this->get_rgba( $color );
		$rgb   = $this->get_rgb( $rgba );
		$light = array(
			min( 255, ( ( $amount / 100 ) * 255 ) + $rgb[0] ),
			min( 255, ( ( $amount / 100 ) * 255 ) + $rgb[1] ),
			min( 255, ( ( $amount / 100 ) * 255 ) + $rgb[2] ),
		);
		$light = array_merge( $light, array( $rgba[3] ) );
		return $this->format( $light, $format );
	}

	/**
	 * Darken color by a given amount.
	 *
	 * @param mixed  $color the color.
	 * @param float  $amount the darkening factor (0-100 ideally).
	 * @param string $format the format of the result.
	 */
	public function darken( $color, $amount = 15, $format = 'array' ) {
		$rgba = $this->get_rgba( $color );
		$rgb  = $this->get_rgb( $rgba );
		$dark = array(
			max( 0, ( ( 100 - $amount ) / 100 ) * $rgb[0] ),
			max( 0, ( ( 100 - $amount ) / 100 ) * $rgb[1] ),
			max( 0, ( ( 100 - $amount ) / 100 ) * $rgb[2] ),
		);
		$dark = array_merge( $dark, array( $rgba[3] ) );
		return $this->format( $dark, $format );
	}

	/**
	 * Reverse color.
	 *
	 * @param mixed  $color the color.
	 * @param string $format the result format.
	 */
	public function reverse( $color, $format = 'array' ) {
		$rgba = $this->get_rgba( $color );
		$rgb  = $this->get_rgb( $rgba );
		$rev  = array(
			255 - $rgb[0],
			255 - $rgb[1],
			255 - $rgb[2],
		);
		$rev  = array_merge( $rev, array( $rgba[3] ) );
		return $this->format( $rev, $format );
	}

	/**
	 * Shade color by a given amount (darken for light color, lighten dark color).
	 *
	 * @param mixed  $color the color.
	 * @param float  $amount the shading factor (0-100).
	 * @param string $format the desired format for the result.
	 */
	public function shade( $color, $amount = 15, $format = 'array' ) {
		$rgba    = $this->get_rgba( $color );
		$rgb     = $this->get_rgb( $rgba );
		$is_dark = $this->is_dark( $rgb );
		if ( $is_dark ) {
			return $this->lighten( $rgba, $amount, $format );
		} else {
			return $this->darken( $rgba, $amount, $format );
		}
	}

	/**
	 * Accentuate color (lighten light color, darken dark).
	 *
	 * @param mixed  $color the color.
	 * @param float  $amount the accentuation factor (0-100).
	 * @param string $format the desired format for the result.
	 */
	public function accent( $color, $amount = 15, $format = 'array' ) {
		$rgba    = $this->get_rgba( $color );
		$rgb     = $this->get_rgb( $rgba );
		$is_dark = $this->is_dark( $rgb );
		if ( $is_dark ) {
			return $this->darken( $rgba, $amount, $format );
		} else {
			return $this->lighten( $rgba, $amount, $format );
		}
	}

	/**
	 * Rgb to gray scale.
	 *
	 * @param mixed  $color the color.
	 * @param string $format the desired format for the result.
	 */
	public function grayscale( $color, $format = 'array' ) {
		$rgba = $this->get_rgba( $color );
		$rgb  = $this->get_rgb( $rgba );
		$av   = ( $rgb[0] * 21 + $rgb[1] * 72 + $rgb[2] * 7 ) / 100;
		return $this->format( array( $av, $av, $av, $rgba[3] ), $format );
	}

	/**
	 *  Checks if is dark color.
	 *
	 * @param mixed $color the color.
	 */
	public function is_dark( $color ) {
		$rgba = $this->get_rgba( $color );
		$rgb  = $this->get_rgb( $rgba );
		return ( ( $rgb[0] * 33 + $rgb[1] * 50 + $rgb[2] * 16 ) / 25500 ) < 0.5;
	}

	/**
	 * Returns the unique instance of this class.
	 */
	public static function instance() {
		if ( null === self::$instance ) {
			self::$instance = new self();
		}
		return self::$instance;
	}
}
